open Interval ;;

(********************************
  alternative box design: PARTIAL function from variables to intervals (i.e., association list, 
     for which library provides functions)
*)

type box = { mutable bounds:interval array ; mutable empty: bool} ;;

(* constructors *)

let is_consistent (b: box)= 
  let is_consistent { lo= lo; hi= hi } = lo<=hi in
  b.empty or 
    let i= ref 0 in
    while !i<Array.length b.bounds && is_consistent b.bounds.(!i) do
      i:= !i+1
    done;
    !i=Array.length b.bounds

let empty_box () = { bounds= Array.make 0 { lo= 0.; hi= 0.}; empty= true }

let zero_dim () = { bounds= (Array.of_list []); empty= false }

let make (bd: interval array) =
  let result= ({ bounds= Array.map Interval.copy bd; empty= false }: box)
  in begin
    assert (is_consistent result);
    result
  end

let init (dim: int) (f: int -> interval) =
  let fp k = Interval.copy (f k) in
  { bounds= Array.init dim fp; empty= false }

let full_box (n: int) = 
  let f k = everything () in
  init n f

let is_empty (b: box)= b.empty

let dim (b: box)= 
  if is_empty b then -1
  else Array.length b.bounds


let copy { bounds= bounds; empty= empty}= 
  { bounds= Array.map Interval.copy bounds; empty= empty }


(* updates #1 by copying certain bounds of the box #2 to the corresponding bounds of box #1 which corresponds to all the variables in the list #3 , iff the dimension of the box #2 is greater than variable value in the var_list #3*)

let copy_box_bounds (oldbox : box) (box : box) (var_list : int list) : unit=
    List.iter (function v -> (if ((dim box) > v) && (box.bounds.(v).lo <> neg_infinity) then oldbox.bounds.(v)<-Interval.copy box.bounds.(v))) var_list
(*
let rec copy_box_bounds (oldbox : box) (box : box) (var_list : int list) : unit =
match var_list with
[]-> ()
| first::rest ->
                  oldbox.bounds.(first)<- Interval.copy box.bounds.(first);
                 copy_box_bounds oldbox box rest*)

let empty_box () = { bounds= Array.of_list []; empty= true }

let print (box: box) =
 if is_empty box then
    print_string "empty"
  else begin
    print_string "[ ";
    Array.iter Interval.print box.bounds;
    print_string " ]";
  end

let print_debug (box: box) =
 if is_empty box then
    print_string "empty"
  else begin
    print_string "[ ";
    Array.iter Interval.print_debug box.bounds;
    print_string " ]";
  end


(*let print_boxlist (b:box list) = 
    List.iter print b
*)

(* to print the boxes in the binary format *)

let print_boxlist (b:box list) (str :string)=
  let bx_len = ref 0 in begin
  let print_box_bin (box: box) =
	if is_empty box then
    	   print_string "e"
  	else begin
     	if !bx_len != (List.length b) then
	   print_string str;
    	   bx_len:=!bx_len+1;
  	end;
	Array.iter Interval.print_ivl_bin box.bounds;
in
  List.iter print_box_bin b;
end

let set_bound (b: box) (k: int) (iv: interval) = 
  assert(0<=k && k<(dim b));
  b.empty<-false;
  b.bounds.(k)<- iv


let get_bound (b: box) (k: int) = 
  assert(0<=k && k<(dim b));
  b.bounds.(k)

let cart_prod (b: box list): box= 
  if List.exists is_empty b then
    empty_box ()
  else
    make (Array.concat ((List.map (function b -> b.bounds) b)))

let cart_prod2 (b1: box) (b2: box): box= 
  if is_empty b1 or is_empty b2 then
    empty_box ()
  else 
    make (Array.append b1.bounds b2.bounds)

let cart_prod2_shared (b1: box) (b2: box)= 
  if is_empty b1 or is_empty b2 then
    empty_box ()
  else 
    { bounds = (Array.append b1.bounds b2.bounds); empty= false }

let extend (b: box) (iv: interval)= 
  cart_prod2 b { bounds= Array.make 1 iv; empty= false} 

let set_empty (b: box) =
  b.bounds<-(Array.of_list []);
  b.empty<-true

let set_full (b: box) =
  for i= 0 to (dim b)-1 do
    b.bounds.(i)<- everything ()
  done;
  b.empty<-false

let volume (b: box)=
  if b.empty then 0.0
  else Array.fold_left ( *.) 1.0 (Array.map width b.bounds) 

let contains (b: box) (p: float array ): bool=
  Util.array_for_all_i (fun x i -> Interval.contains (get_bound b i) x) p<0 

let interior (b: box) (p: float array ): bool=
  let interior iv x =  iv.lo < x && x<iv.hi in
  Util.array_for_all_i (fun x i -> interior (get_bound b i) x) p<0 

let rec diff (b1: box) (b2: box): box list =
  if is_empty b1 then [] 
  else if is_empty b2 then [ copy b1 ]
  else if (dim b1)=0 then []
  else begin
    assert ((dim b1)=(dim b2));
    let rest_b1= make (Array.sub b1.bounds 0 ((dim b1)-1)) 
    and rest_b2= make (Array.sub b2.bounds 0 ((dim b2)-1))
    and first_b1= b1.bounds.((dim b1)-1)
    and first_b2= b2.bounds.((dim b2)-1)
    in let sides = List.map (extend rest_b1) (Interval.diff first_b1 first_b2 ) 
    and center = List.map (function b -> extend b first_b2) (diff rest_b1 rest_b2)
    in List.filter (function b -> not (is_empty b)) (List.append sides center)
  end

let rec diff_under_appr (b1: box) (b2: box): box list =
  if is_empty b1 then [] 
  else if is_empty b2 then [ copy b1 ]
  else if (dim b1)=0 then []
  else begin
    assert ((dim b1)=(dim b2));
    let rest_b1= make (Array.sub b1.bounds 0 ((dim b1)-1)) 
    and rest_b2= make (Array.sub b2.bounds 0 ((dim b2)-1))
    and first_b1= b1.bounds.((dim b1)-1)
    and first_b2= b2.bounds.((dim b2)-1)
    in let sides = List.map (extend rest_b1) (Interval.diff first_b1 first_b2 ) 
    and center = List.map (function b -> extend b first_b2) (diff rest_b1 rest_b2)
    in List.filter (function b -> not (is_empty b)) (List.append sides center)
  end

let width (b: box): float =
  if is_empty b then
    0.0
  else
    List.fold_left max neg_infinity (List.map Interval.width (Array.to_list b.bounds))

let diagonal (b: box): float =
  if is_empty b then
    0.0
  else
    let sqr x = x *. x in 
    sqrt (List.fold_left (+.) 0.0 (List.map (Util.compose sqr Interval.width) (Array.to_list b.bounds)))

let union_d (b: box) (bu: box) = 
  assert (is_empty b || is_empty bu || (dim b)=(dim bu));
  if not (is_empty bu) then 
    if is_empty b then begin
      b.empty<-false;
      b.bounds<-bu.bounds
    end else 
      for i= 0 to (dim b)-1 do
	b.bounds.(i)<-union b.bounds.(i) bu.bounds.(i)
      done

(* inefficient implementation!!! *)
let union2 (b1: box) (b2: box) = 
  assert (is_empty b1 || is_empty b2 || (dim b1)=(dim b2));
  if (is_empty b2) then b1
  else if (is_empty b1) then b2 
  else 
    make (Array.of_list (List.map2 union (Array.to_list b1.bounds) (Array.to_list b2.bounds)))

let intersection2 (b1: box) (b2: box): box = 
  assert (is_empty b1 || is_empty b2 || (dim b1)=(dim b2));
  if (is_empty b1) || (is_empty b2) then empty_box ()
  else 
    let res= (List.map2 Interval.intersect (Array.to_list b1.bounds) (Array.to_list b2.bounds)) in
    if List.exists Interval.is_empty res then
      empty_box ()
    else make (Array.of_list res)

let union (b: box list) = 
  List.fold_left union2 (empty_box ()) b

let widest_var ({bounds= bounds; empty= empty}: box) (k: int ) =
  assert(not empty);
  let wvar = ref (-1)
  in let wwidth = ref neg_infinity
  in
  (*for i= 0 to (Array.length bounds)-1 do*)
   for i= 0 to (k-1) do
    if Interval.width (bounds.(i)) >= !wwidth then begin
       wvar:= i ;
      wwidth:= Interval.width (bounds.(i))
    end
  done ;
  !wvar

let bisect (b: box) (k: int)=
  assert(not b.empty);
  let b1= copy b
  and b2= copy b
  and (iv1, iv2)= bisect(b.bounds.(k))
  in begin
    set_bound b1 k iv1;
    set_bound b2 k iv2;
    assert(is_consistent b1);
    assert(is_consistent b2);
    (b1, b2)
  end

let bisection_face (b: box) (k: int)=
  assert(not b.empty);
  let b1= copy b in
  set_bound b1 k (float_to_intv (midpoint (get_bound b1 k)));
  b1

let midpoint_box (b: box): box= 
  assert(not (is_empty b));
  let initf i = float_to_intv (midpoint (get_bound b i)) in
  { bounds= Array.init (dim b) initf; empty= false }

let midpoint (b: box): float array= 
  assert(not (is_empty b));
  Array.init (dim b) (fun i -> midpoint (get_bound b i)) 

let copy_bounds (b1:box) (b2:box) =
  assert(not (is_empty b2));
  b1.empty<-false;
  b1.bounds<- Array.copy b2.bounds

let proj1 (b: box) =
  assert(not (is_empty b));
  make (Array.sub b.bounds 0 ((dim b)-1))

let proj (b: box) (from: int) (till: int) = 
  if (is_empty b) then
    b
  else
    make (Array.sub b.bounds from till)


let rec blowup (b: box) (k: int): box list =
  assert(not (is_empty b)); 
  if((dim b)=0) then [ zero_dim () ]
  else 
    let res= Util.prod extend (blowup (proj1 b) k) (Array.to_list (split b.bounds.((dim b)-1) k)) in begin
      assert(union(res)=b);
      res
    end

let box_precision (b:box) (d:float) =
  for i =0 to ((dim b)-1) do
     b.bounds.(i) <- Interval.precision b.bounds.(i) d
  done

(*let box_precision (b:box) (d:float) =
  for i =0 to ((dim b)-1) do
     b.bounds.(i) <- {lo=float_of_int (int_of_float ( b.bounds.(i).lo  *. d));hi = float_of_int (int_of_float ( b.bounds.(i).hi  *. d))}  
  done*)

(*
let rec pow x y =
     if y <> 0 then begin 
        x* pow x (y-1)
     end
     else
       1
*)

let diff_width_rel (b1:box) (b2:box) (eps:float): bool =
 if (is_empty b2) then
   true
 else
   let i= ref 0 in
   while  !i<(dim b1) && (Interval.width b1.bounds.(!i)=0.0 || ((Interval.width b2.bounds.(!i)) /. (Interval.width b1.bounds.(!i))) >= (1.0 -. eps)) do
     i:=!i+1;
   done;
   !i<(dim b1)

let diff_width_rel_vars (b1: box) (b2: box) (eps:float): int list =
  assert((dim b1)=(dim b2));
  let res= ref [] in
  for i= 0 to (dim b1-1) do
    if (((Interval.width b2.bounds.(i)) /. (Interval.width b1.bounds.(i))) < (1.0 -. eps)) then 
      res:= i::!res;
  done;
  !res

let diff_vars (b1: box) (b2: box): int list =
  assert((dim b1)=(dim b2));
  let res= ref [] in
  for i= 0 to (dim b1-1) do
    if not (b1.bounds.(i)=b2.bounds.(i)) then 
      res:= i::!res;
  done;
  !res

let compare_box (b1:box) (b2:box) (i:float) =
  let b11 = copy b1 in 
  let b22 = copy b2 in
  box_precision b11 i;  
  box_precision b22 i;
  b11 <> b22 

let cmp_bx (b1:box) (b2:box)  =
   let list = ref [] in 
     for i =0 to (dim b1)-1 do
       if Interval.equal b1.bounds.(i) b2.bounds.(i) then
          list:=!list @ [i] 
     done;
    !list

let equal (b1: box) (b2: box): bool =
  if (is_empty b1) && (is_empty b2) then true
  else if dim b1!=dim b2 then false else
  let i= ref 0 in
  while (!i<dim b1) && Interval.equal b1.bounds.(!i) b2.bounds.(!i) do
    i:= !i+1
  done;
  !i>=dim b1

let subset (b1: box) (b2: box): bool =
  if (is_empty b1) then 
    true
  else if (is_empty b2) then
    false 
  else begin
    assert(dim b1=dim b2);
    let i= ref 0 in
    while (!i<dim b1) && Interval.subset b1.bounds.(!i) b2.bounds.(!i) do
      i:= !i+1
    done;
    !i>=dim b1
  end
 
let vrml (b: box list) (red: int) (green: int) (blue: int) =
  let vrml_box (b: box) =
    let print_color () =
      print_string "diffuseColor ";
      print_int red;
      print_string " ";
      print_int green;
      print_string " ";
      print_int blue;
      print_endline "";
    and print_trans () = 
      let print_left (iv: interval) = 
	print_float ((Interval.width iv) +. (iv.lo *. 2.));
	print_string " " in
      print_string "translation ";
      Array.iter print_left b.bounds;
      for i= (dim b)+1 to 3 do 
	print_string " 1. "
      done;
      print_endline ""
    and print_scale () = 
      let print_width (iv: interval) = 
	print_float ((Interval.width iv)); 
	print_string " " in
      print_string "scale ";
      Array.iter print_width b.bounds;
      for i= (dim b)+1 to 3 do 
	print_string " 1 "
      done;
      print_endline "" in
    print_endline "Transform { ";
    print_scale ();
    print_trans ();
    print_endline "children [
                    Shape {
                    appearance Appearance {
                    material Material {";
    print_color ();
    print_endline "transparency 0.1
                 }
               }
               geometry Box {";
    print_string "}
}
  ] 
}
" in
  List.iter vrml_box b

let intersecting (b1: box) (b2: box): bool =
  if (is_empty b1) || (is_empty b2) then false
  else (Util.array_for_all2 Interval.intersecting b1.bounds b2.bounds<0)

let overlapping (b1: box) (b2: box): bool =
  if (is_empty b1) || (is_empty b2) then false
  else (Util.array_for_all2 Interval.overlapping b1.bounds b2.bounds<0)
