{
open Parser        (* The type token is defined in parser.mli *)

let ignore_lex lb = ignore (Lexing.lexeme lb)

exception Unbalanced_Comment
exception Eof


}

let WHITESPACE = [' ' '\t' '\n' '\r'] 
let VARIABLE = ['a'-'z''A'-'Z']['a'-'z''A'-'Z''0'-'9''_''\'']*   

let FLOAT = ['0'-'9']+ ('.'['0'-'9']*)? (('e'|'E')('+'|'-')?['0'-'9']+)?
let IVFLOAT = ( '-'? ['0'-'9']+ ('.'['0'-'9']*)? (('e'|'E')('+'|'-')?['0'-'9']+)? ) |"INF"|"-INF"

rule token = parse
          | '#'['0'-'9']*    { let bufstring= (Lexing.lexeme lexbuf) in
	                       BOOLVAR(int_of_string(String.sub bufstring 1 ((String.length bufstring)-1))) }
          | "TRUE"         { TRUE }
          | "FALSE"        { FALSE } 
          | "EXISTS"       { EXISTS } 
          | "EXISTS*"      { EXISTS_NOBRANCH } 
          | "FORALL"       { FORALL } 
          | "FORALL*"      { FORALL_NOBRANCH } 
          | "/\\"          { AND }
          | "\\/"          { OR }
          | "==>"          { IMPLIES }
          | "="            { EQ } 
          | "<>"           { NE } 
          | "<="           { LE } 
          | "<"            { LT } 
          | ">="           { GE } 
          | ">"            { GT } 
          | "PSD"          { POSSEMDEF }
          | "SUBSET"       { SUBSET }
          | "SIN"          { SIN }
          | "COS"          { COS }
          | "TAN"          { TAN }
          | "EXP"          { EXP }
          | "LOG"          { LOG }
          | "ASIN"         { ASIN }
          | "ACOS"         { ACOS }
          | "ATAN"         { ATAN }
          | '+'            { PLUS }
          | '-'            { MINUS }
          | '*'            { TIMES }
          | '/'            { DIV }
          | "^2"           { SQR }
          | "^"            { POW }
          | "0"            { ZERO }
          | WHITESPACE    { token lexbuf }     (* skip whitespaces *)
          | "PI"           { INTERVAL({Interval.lo= Constants.pi_lo; Interval.hi= Constants.pi_hi }) }
          | "E"            { INTERVAL({Interval.lo= Constants.e_lo; Interval.hi= Constants.e_hi }) }
          | "INF"           { FLOAT("infinity") }
          | "-INF"           { FLOAT("neg_infinity") }
          | "+INF"           { FLOAT("infinity") }
          | FLOAT { FLOAT(Lexing.lexeme lexbuf) }
          | '[' WHITESPACE* IVFLOAT WHITESPACE* ',' WHITESPACE* IVFLOAT WHITESPACE* ']' { 
	                       let drop_leading_space (s: string): string =
				 let i= ref 0 in
				 while s.[!i]=' ' do
				   i:= !i+1
				 done;
				 String.sub s !i ((String.length s)- !i) 
			       and drop_trailing_space (s: string): string =
				 let i= ref ((String.length s)-1) in
				 while s.[!i]=' ' do
				   i:= !i-1
				 done;
				 String.sub s 0 (!i+1) in
	                       let bufstring= (Lexing.lexeme lexbuf) in
			       let comma= String.index bufstring ',' in
			       let lostr= drop_trailing_space(drop_leading_space(String.sub bufstring 1 (comma-1)))
                               and histr= drop_trailing_space(drop_leading_space(String.sub bufstring (comma+1) ((String.length bufstring)-comma-2))) in 
			       let lo= 
				 if (String.length lostr)>=4 && (String.sub lostr 0 4)="-INF" then 
				   neg_infinity 
				 else 
				   float_of_string(lostr) in
			       let hi= 
				 if  (String.length histr)>=3 && (String.sub histr 0 3)="INF" then 
				   infinity 
				 else 
				   float_of_string(histr) in 
			       INTERVAL({Interval.lo= lo; Interval.hi= hi})}
          | VARIABLE   { VARIABLE(Lexing.lexeme lexbuf) }
          | '('            { LPAREN }
          | ')'            { RPAREN }
	  | '['            { LBRACKET }
	  | ']'            { RBRACKET }
	  | ','            { COMMA }
          | ';'            { END }
          | "@R"           { REDUNDANT }
          | eof            { raise Eof }
          | "(*"   { ignore_lex lexbuf ; comment lexbuf ; token lexbuf }
          | "*)"   { raise Unbalanced_Comment }
(*          |  _     { traite_token lexbuf ; token lexbuf } *)
          and comment = parse
            "(*"   { ignore_lex lexbuf ; comment lexbuf ; comment lexbuf }
          |  "*)"   { ignore_lex lexbuf }
          |   _     { ignore_lex lexbuf ;  comment lexbuf }
          |   eof   { raise Unbalanced_Comment } 
