/* #include "csdp/lib/declarations.h" */
#include "dsdp5.h"
#include <caml/alloc.h>
#include <caml/mlvalues.h>
#include <caml/memory.h>
#include "assert.h"
#include <stdio.h>
#include <math.h>

#define TO_INTREF(x) ((INTERVAL *)&Field(x, 0))
#define TO_INTV(x) (*(TO_INTREF(x)))
#define ALLOC_INTV alloc_small(sizeof(double)*2, Double_array_tag);

#define TRUE 1
#define FALSE 0

#define INFINITY 1.0/0.0

//#define DEBUG

/* 
   three LMIs:
   * a dense matrix with entries ((y1, ..., yk), (y_k+1, ... y_2k), ... ) is PSD
   * a diagonal matrix with lower bounds on y1, ..., y_kk
   * a diagonal matrix with upper bounds on y1, ..., y_kk

   so we have ONE LMI with three blocks

   LMI formulation corresponds to DUAL SDP formulation, but with matrices A negated (by bringing it to the righ-hand side)!

   check: un/symmetric matrices!!!!!!!!!!!!
*/

int DSDP_index(int i, int j) {
  assert(j<=i);
  return i*(i+1)/2+j;
}  


double get_entry(value M, int i, int j, int left) {
#ifdef DEBUG
  //  printf("%d %d %F\n", i, j, Double_field(Field(Field(M, i), j), left?0:1)); fflush(stdout);
#endif
  return Double_field(Field(Field(M, i), j), left?0:1);
}

double get_fentry(value f, int i, int j) {
  //#ifdef DEBUG
  //  printf("F: %d %d %F\n", i, j, Double_field(Field(f, i), j)); fflush(stdout);
  //#endif
  return Double_field(Field(f, i), j);
}

void set_fentry(value f, int i, int j, double x) {
  assert(Tag_val(f)<No_scan_tag);
  assert(Tag_val(Field(f, i))==Double_array_tag);
  Store_double_field(Field(f, i), j, x);
}

value sdp_solver_() {
  return copy_string("DSDP 5.7 by Steve Benson, Yinyu Ye, and Xiong Zhang");
}

value sdpx_(value M, 
 	   value f /* float matrix */) {
  CAMLparam2(M, f);
  CAMLlocal1(result);

#define NUMSYMENTRIES (Wosize_val(M)*(Wosize_val(M)+1)/2) /* number of entries of symmetric matrix */

  DSDPTerminationReason stop_code;
  int i, j;

  DSDP dsdp;

  double blocksizes[2];

  double C[NUMSYMENTRIES];
  int A_psd_ind[NUMSYMENTRIES]; 
  double A_psd_val[NUMSYMENTRIES]; 
  int A_1_ind[Wosize_val(M)][2]; /* block 1, indices, sparse with 2 entries */
  double A_1_val[Wosize_val(M)][2]; /* block 1, values, sparse with 2 entries */

  SDPCone cone_psd;

  if(DSDPCreate(NUMSYMENTRIES, &dsdp)) {
    assert(0);
  }

  /* set constants of (dual) objective  b */
  for(i= 0; i<Wosize_val(M); i++) { /* each matrix entry */
    for(j= 0; j<=i; j++) {
      if(DSDPSetDualObjective(dsdp, 
			      DSDP_index(i, j)+1 /* indices start with 1 */, 
			      get_fentry(f, i, j))) {
	assert(0);
      }
#ifdef DEBUG
      printf("Obj: %d %F\n", DSDP_index(i, j), get_fentry(f, i, j));
#endif
    }
  }

  if(DSDPCreateSDPCone(dsdp, 1, &cone_psd)) { 
    assert(0);
  }
  //  SDPConeSetBlockSize would improve efficiency
  //  SDPConeSetSparsity would imrpove efficiency

  SDPConeSetZeroMat(cone_psd, 0, 0, Wosize_val(M));
#ifdef DEBUG
  SDPConeViewDataMatrix(cone_psd, 0, 0);
#endif

		   
  /* coefficient matrices A, current matrix is i*Wosize_val(M)+j */
  for(i= 0; i<Wosize_val(M); i++) { 
    for(j= 0; j<=i; j++) { 
      A_psd_ind[DSDP_index(i, j)]= DSDP_index(i, j);
      A_psd_val[DSDP_index(i, j)]= -1.0;
      if(SDPConeSetSparseVecMat(cone_psd, 
				0 /* block */, 
				DSDP_index(i, j)+1 /* variable, 0 refers to C */,
				Wosize_val(M) /* matrix dimension */,
				0 /* starting index */,
				&(A_psd_ind[DSDP_index(i, j)]),
				&(A_psd_val[DSDP_index(i, j)]),
				1 /* array length */)) {
	assert(0);
      }
#ifdef DEBUG     
      SDPConeViewDataMatrix(cone_psd, 0, DSDP_index(i, j)+1);
#endif
    }
  }

  /* now create cone that enforces bounds */
  LPCone cone_bounds;
  if(DSDPCreateLPCone(dsdp, &cone_bounds)) { 
    assert(0);
  }

  int bounds_nnzin[NUMSYMENTRIES+2]; /* specifies columns */
  bounds_nnzin[0]= 0;
  for(i= 1; i<NUMSYMENTRIES+1; i++) { /* two entries in each column (lower bounding constraint, upper bounding constraint), all in last column (right-hand sides are viewed as additional row */
    bounds_nnzin[i]= bounds_nnzin[i-1]+2; 
  }
  bounds_nnzin[NUMSYMENTRIES+1]= bounds_nnzin[NUMSYMENTRIES] + 2*NUMSYMENTRIES;

  int bounds_row[4*NUMSYMENTRIES];  /* one right-hand side, one left-hand side for lower bounds and upper bounds */
  double bounds_val[4*NUMSYMENTRIES];
  double lower_bound, upper_bound; /* lower bound and upper bound to all the entries, to increase DSDP performance */

  lower_bound= INFINITY; 
  upper_bound= -INFINITY;
  for(i= 0; i<Wosize_val(M); i++) { /* rows */
    for(j= 0; j<=i; j++) { /* columns */
      bounds_row[2*DSDP_index(i, j)]= DSDP_index(i, j);    /* left-hand side, lower bound */
      bounds_val[2*DSDP_index(i, j)]= -1.0;
      
      bounds_row[2*DSDP_index(i, j)+1]= DSDP_index(i, j)+NUMSYMENTRIES;  /* left-hand side, upper bound */
      bounds_val[2*DSDP_index(i, j)+1]= 1.0; 

      double hi_val, low_val, val1, val2;
      val1= get_entry(M, i, j, 1);
      val2= get_entry(M, j, i, 1);
      low_val= val1>val2?val1: val2;
      if(low_val<lower_bound) lower_bound= low_val;

      bounds_row[2*NUMSYMENTRIES+DSDP_index(i, j)]= DSDP_index(i, j);
      bounds_val[2*NUMSYMENTRIES+DSDP_index(i,j)]= -low_val; /* right-hand side, lower bound */

      val1= get_entry(M, i, j, 0);
      val2= get_entry(M, j, i, 0);
      hi_val= val2>val1?val1: val2;
      if(hi_val>upper_bound) upper_bound= hi_val;
  
      bounds_row[3*NUMSYMENTRIES+DSDP_index(i, j)]= NUMSYMENTRIES+DSDP_index(i, j);
      bounds_val[3*NUMSYMENTRIES+DSDP_index(i, j)]= hi_val; /* right-hand side, upper bound */

      //      DSDPSetY0(dsdp, DSDP_index(i, j)+1, (hi_val+low_val)/2.0); /* center as initial value, only to increase efficiency; might not be PSD, so not sure whether this works */
    }
  }

  if(LPConeSetData2(cone_bounds, 
		    2*NUMSYMENTRIES, /* two bounds for each variable */
		    bounds_nnzin,
		    bounds_row,
		    bounds_val)) {

    assert(0);
  }
#ifdef DEBUG
  LPConeView(cone_bounds);
#endif

  /* setting parameters */
  //  DSDPSetPenaltyParameter(dsdp, 1e24); /* this sometimes avoids that DSDP wrongly detects unbounded solution set */
  DSDPSetMaxIts(dsdp, 200);
  //  DSDPBoundDualVariables(dsdp, 2*lower_bound-upper_bound, 2*upper_bound-lower_bound); /* set bounds on the variables with safety margin, problem: then we sometimes get error msg: unbounded solution set; this increases run-time on henrion.rs, so currently we do not use it */

#ifdef DEBUG
  DSDPSetStandardMonitor(dsdp, 1);
#endif


  /* solving problem */
  if(DSDPSetup(dsdp)) {
    assert(0);
  }

  //  DSDPSetScale(dsdp, 1e-12);
  //    DSDPSetScale(dsdp, 0.1);  /* default */

  if(DSDPSolve(dsdp)) {
    assert(0);
  }

  DSDPStopReason(dsdp, &stop_code);
  
#ifdef DEBUG
  printf("SDP stop code: %d\n", stop_code);
#endif

  double solution[NUMSYMENTRIES];
  //  double *solution;
  int solution_size;
  //  SDPConeGetXArray(cone_psd, 0, &solution, &solution_size);
   //  solution_size= Wosize_val(M);
   //  SDPConeComputeX(cone_psd, 0, solution, solution_size);

  DSDPGetY(dsdp, solution, NUMSYMENTRIES);

  #ifdef DEBUG
  for(i= 0; i<NUMSYMENTRIES; i++) {
      printf("Sol: %d %F\n", i, solution[i]);
  }
  printf("\n");
  #endif

  /*
  double solutionY[NUMSYMENTRIES];
  double solutionS[NUMSYMENTRIES];
  SDPConeComputeS(cone_psd, 0, solutionY, NUMSYMENTRIES, solutionS, Wosize_val(M));
  */

  for(i= 0; i<Wosize_val(M); i++) {
    for(j= 0; j<Wosize_val(M); j++) {
      if(j<=i) {
	set_fentry(f, i, j, solution[DSDP_index(i, j)]);
      } else
	set_fentry(f, i, j, solution[DSDP_index(j, i)]);
    }
      //      assert(get_fentry(f, i, j)==solution1[i*Wosize_val(M)+j]);
  }

  if(DSDPDestroy(dsdp)) {
    assert(0);
  }

#define SDPSuccess 0
#define SDPUnbounded 1
#define SDPNotFeasible 2
#define SDPNoSuccess 3
  if(stop_code==DSDP_CONVERGED) { /* optimum found */
    result= Val_int(SDPSuccess);
  } else if(stop_code==DSDP_INFEASIBLE) { /* dual unbounded */
    result= Val_int(SDPNotFeasible);
  } else if(stop_code==DSDP_UNBOUNDED) { /* primal infeasible */
    result= Val_int(SDPUnbounded);
  } else {
    result= Val_int(SDPNoSuccess);
  }
  
  #ifdef DEBUG
  //  DSDP_EventLogSummary();

  fflush(0);
  #endif
  
  CAMLreturn(result);

  /* the following was an attempt at an CSDP interface, but I could not make it work, since CSDP depends on FORTRAN libraries that I could not make run together with OCaml */  
#if 0
  unsigned int n;
  unsigned int k; 
  struct blockmatrix C;
  double *a;
  struct constraintmatrix *constraints;
  struct blockmatrix *pX0;
  double **py;
  struct blockmatrix *pZ0;

  struct blockrec possemdef_block;
  struct blockrec bounds_block;

  int i,j;

  n= Wosize_val(M) + 2*Wosize_val(M)*Wosize_val(M); /* interval matrix + entries for constraining bounds  */
  k= 2*Wosize_val(M)*Wosize_val(M); /* number of constraints= 2*number of matrix entries  */

  alloc(a, k+1); /* right-hand side of constraints, 0-th entry not used! */
  for(i= 0; i<Wosize_val(M); i++) {
    for(j= 0; j<Wosize_val(M); j++) {
      a[2*(Wosize_val(M)*i+j)+1]= get_entry(M, i, j, TRUE); /* left bound */ 
      a[2*(Wosize_val(M)*i+j)+2]= get_entry(M, i, j, FALSE); /* right bound */
    }
  }

  possemdef_block.blockcategory= MATRIX;
  bounds_block.blockcategory= DIAG;
  for(i= 0; i< 2*Wosize_val(M); i++) { /* objective does not contain bounds variables */
    bounds_block.data.vec[i]= 0.0;
  }
  for(i= 0; i< 2*Wosize_val(M); i++) { /* all entries zero except for v */
    for(i= 0; i< 2*Wosize_val(M); i++) { 
      if(i==Val_int(Field(v, 0)) && j==Val_int(Field(v, 1))) {
	possemdef_block.data.vec[ijtok(i,j,Wosize_val(M))]= 0.0;
      } else {
	possemdef_block.data.vec[ijtok(i,j,Wosize_val(M))]= Bool_val(d) ? 1.0:-1.0;
      }
    }
  }
  /*
  initsoln(n, 
	   k, 
	   C, 
	   a,
	   constraints,
	   pX0,
	   py,
	   pZ0);
  /*  easy_sdp();*/
#endif
}
